
#ifndef _GROWABLEARRAY_H_
#define _GROWABLEARRAY_H_

#include <stdexcept>

/** \file

   <b>Copyright and Use</b>

   You may use this source code without limitation and without
   fee as long as you include:

<pre>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2001, 2002, 2003
</pre>

   This software is provided "as is", without any warrenty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.

   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>

 */


/**
   GrowableArray template is used to create array objects that will
   grow as elements are added to the end of the array.  This is
   similar to the STL &lt;vector&gt; class.

   This array class is designed for dense arrays where the all
   elements of the array are used.

   Usage:

<ul>
<li>
<p>
       Elements are added to the end of the array via the append
       function.
</p>
</li>
<li>
<p>
       Elements in the array can be accessed via the [] operator.
</p>
</li>
<li>
<p>
       If the elements in the array are dynamically allocated, the
       user is responsible for deallocating these elements.  For
       example, if the GrowableArray template is used to define
       an array of pointers to an object, the user is responsible for
       deallocating the objects referenced by the pointers.  The
       GrowableArray destructor will only deallocate the array
       itself.
</p>
</li>
</ul>

  A doubling algorithm is used when the data size is expanded because
  it minimizes the amount of copying that must be done.  The array
  will quickly grow to a the size needed to accomodate the data set
  and no more copying will be necessary.  For large arrays there is the
  drawback that more memory may be allocated than is needed, since
  the amount of memory used grows exponentially.

 */
template <class T>
class GrowableArray {
private:
  typedef enum { StartArraySize = 64 } bogus;
  size_t num_elem;     // number of data elements in the array
  size_t array_size;   // array size (e.g., data capacity)
  T *pArray;

private:
  /**
     Double the amount of memory allocated for the array.
     Return true if memory allocation succeeded, false
     otherwise.
  */
  bool twice()
  {
    bool rslt;

    T *old_array = pArray;
    size_t new_size = array_size * 2;
    
    pArray = new T [ new_size ];
    if (pArray != 0) {
      rslt = true;
      for (int i = 0; i < array_size; i++) {
        pArray[i] = old_array[i];
      }
    
      delete [] old_array;
    
      array_size = new_size;
    }
    else {
      rslt = false;
    }

    return rslt;
  } // twice


  void init(size_t initialSize )
  {
    array_size = StartArraySize;
    while (initialSize > array_size) {
      array_size = array_size * 2;
    }
    pArray = new T[ array_size ];
    num_elem = initialSize;
  } // init

public:
  GrowableArray()
  {
    init( 0 );
  } // GrowableArray constructor


  GrowableArray(size_t initialSize)
  {
    init( initialSize );
  }

  ~GrowableArray()
  {
    if (pArray != 0) {
      delete [] pArray;
    }
  } // GrowableArray destructor

  const size_t length(void) const { return num_elem; }

  T &operator[](const size_t i)
    throw(std::out_of_range)
  {
    if (i >= num_elem) {
      const char *errMsg = "1. GrowableArray::[] - index out of bounds";
      throw std::out_of_range(errMsg);
    }
    return pArray[ i ];
  }

  T operator[](const size_t i ) const
    throw(std::out_of_range)
  {
    if (i >= num_elem) {
      const char *errMsg = "2. GrowableArray::[] - index out of bounds";
      throw std::out_of_range(errMsg);
    }
    return pArray[ i ];
  }

  const T *getData() const { return pArray; }

  /** append an item to the end of the array */
  void append( T item )
    throw(std::runtime_error)
  {

    if (num_elem == array_size) {
      bool allocOK = twice();
      if (!allocOK) {
	const char *errMsg = "GrowableArray::append - memory allocation error";
	throw std::runtime_error( errMsg );
      }
    }

    pArray[ num_elem ] = item;
    num_elem++;
  } // append


  /**
     Expand the number of array data slots by "amount" elements.
     Note that "array_size" is the total amount of storage available
     for data slots.  "num_elem" is the number of data slots.
     The bounds over which the array can be indexed is governed
     by num_elem.  Note that after expand() is called the new
     data elements can be read, but their value is undefined until
     they are initialized.

     This function calls twice(), which copies the old elements
     into the new array.  This function is not called by the
     constructor which takes a size argument, because there is
     not existing array at construction time.
  */
  void expand( size_t amount )
    throw(std::runtime_error)
  {
    bool allocOK = true;

    while (allocOK && num_elem + amount >= array_size) {
      allocOK = twice();
      if (!allocOK) {
	const char *errMsg = "GrowableArray::expand - memory allocation error";
	throw std::runtime_error( errMsg );
      }
    }
    num_elem += amount;
  } // expand


  /** Remove one item from the end of the array. */
  void remove(void)
  {
    if (num_elem > 0)
      num_elem--;
  } // remove

  /**
     Set the number of data elements in the array to
     a new value (note that this will usually be
     smaller than the array size, unless a power of
     two is chosen for "new_size").
  */
  void set_size( size_t new_size )
  {
    if (new_size <= array_size) {
      num_elem = new_size;
    }
    else { // new_size > array_size
      size_t num_new_elem = new_size - num_elem;
      expand( num_new_elem );
    }
  } // set_size

}; // GrowableArray


#endif
