
#ifndef RCARRAY_H
#define RCARRAY_H

/** \file

   <b>Copyright and Use</b>

   You may use this source code without limitation and without
   fee as long as you include:

<pre>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2001, 2002, 2003
</pre>

   This software is provided "as is", without any warrenty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.

   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>

 */

#include "GrowableArray.h"
#include "RCObject.h"
#include "RCPtr.h"
#include "RCBase.h"

/**
   A reference counted, growable array template.

   Objects that are instantiated with this template have three
   characteristics:
<ol>
   <li>
   <p>
      An assignment of the object increments the reference count
      rather than copying memory.  For example, in the code below
      myObj1 is initialized and then assigned to myObj2.  myObj1
      and myObj2 now share the same memory (there was no copy)
      and the reference count will be 2.
   </p>
   <pre>
    #include &lt;iostream&gt;
    #include "RCArray.h"
    
    typedef RCArray<int> IntVec;
    
    main()
    {
      IntVec myObj1, myObj2;
    
      for (int i = 1; i %lt;= 10; i++) {
        myObj1.append( i );
      }
      myObj2 = myObj1;
      for (i = 0; i %lt; 10; i++) {
        std::cout %lt;%lt; myObj2[i] %lt;%lt; " ";
      }
      std::cout %lt;%lt; std::endl;
    }
   </pre>
   <p>
   If this code is complex and executed it will print:
   </p>
   <pre>
   1 2 3 4 5 6 7 8 9 10
   </pre>
   </li>
   <li>
   <p>
      A change to an object with multiple references results in
      a copy-on-write.  That instance of the object will be changed
      while the others remain unchanged.  Since a unique copy has
      been made, the reference count will be reduced by one.
      For example, the assignment
   </p>
   <pre>
         myObj1[4] = 42;
   </pre>
   <p>
      will only change myObj1.  The contents of myObj2 remain
      {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}.
   </p>
   </li>
   <li>
   <p>
      When the reference count reaches zero, the memory for the object
      will be automatically deallocated.  This allows reference
      counted array used as if they were elemental values, without
      worrying about memory allocation and deallocation.
   </p>
   </li>
</ol>

   <p>
   Only the argumentless constructor of RCArray is made public.
   Classes can expose the other constructors and <tt>protected</tt>
   functions like <i>length()</i> and <i>getRefCnt()</i> (which is
   useful in testing and verification) by deriving a class from an
   instance of this template.  For example:
   </p>

   <pre>
    #include "RCArray.h"

    typedef RCArray<double> DoubleArray;

    class DoubleVec : public DoubleArray
    {
    private:
      void init(double intialVal );
    public:
      DoubleVec() : DoubleArray() {}
      DoubleVec(size_t initialSize) : DoubleArray( initialSize ) {}
      DoubleVec(size_t initialSize, double initVal ) : 
        DoubleArray(initialSize, initVal) {}
    
      size_t length() { return DoubleArray::length(); }
      size_t getRefCnt() { return DoubleArray::getRefCnt(); }
    };
   </pre>

   The RCArray template has a single data element, <i>value</i>, which is a
   "smart" pointer.  Note that there is no overloaded assignment
   operator.  The default RCArray assignment copies the <i>value</i>
   element.  This in turn results in a copy of the RCPtr instantiated
   template object (the smart pointer).

   The RCArray template uses the default destructor (oh, the horror!)
   The default destructor will call the destructor on <i>value</i> which
   will, in turn invoke the removeReference() function in RCObject.
   This will decrement the reference count.  When the reference count
   reaches zero, the shared data will be deleted.

   <b>Compiling the code</b>

   The GrowableArray template will throw a runtime exception if
   the array is indexed beyond the bounds of the data.  At least
   in the case of the Microsoft compilers, this requires that 
   you include the -GX flag.  For example:

<pre>
      cl -TP -GX -Zi test.C -o test
</pre>

   (the -Zi flag compiles for debug).

*/
template <class T>
class RCArray : public RCBase<T>
{

protected:
  class SharedData : public RCObject
  {
  public:
    GrowableArray<T> data;

    SharedData() {};
    SharedData(const GrowableArray<T> &v ) { copy( v ); }
    SharedData( size_t initialSize ) : data( initialSize ) {}
    SharedData(const SharedData& rhs) { copy( rhs.data ); }
    ~SharedData() {}
    size_t getRefCnt() { return refCnt(); }

  private:
    void copy(const GrowableArray<T> &v )
    {
      size_t len = v.length();
      for (size_t i = 0; i < len; i++) {
	data.append( v[i] );
      }
    } // copy
  }; // class SharedData

  RCPtr<SharedData> value;

  void init(T intialVal );

  RCArray(size_t initialSize);
  RCArray(size_t initialSize, T initVal );

  T read(const int i) const;
  void write(const int i, const T v);

  size_t length() const;
  size_t getRefCnt();

public:
  RCArray();

  ValRef operator[](const int ix) throw(std::out_of_range);
  T operator[](const int ix) const;

  ValRef operator[](const size_t ix) throw(std::out_of_range);
  T operator[](const size_t ix) const;

  void append(T val);
}; // RCArray


/**
   Initialize the array with a value of type T
 */
template <class T>
inline
void RCArray<T>::init( T initialVal )
{
  size_t len = value->data.length();
  for (size_t i = 0; i < len; i++) {
    value->data[i] = initialVal;
  }
}

/**
   Argumentless constructor: the RCArray will contain no data elements
 */
template <class T>
inline
RCArray<T>::RCArray() : value( new SharedData() ) {}


/**
   Create an RCArray with length() == <i>initialSize</i>.  The length()
   data elements are uninitialized and their value is undefined.
 */
template <class T>
inline
RCArray<T>::RCArray( size_t initialSize ) : 
  value( new SharedData(initialSize) ) {}


/**
   Create an RCArray with length() == <i>initialSize</i>, where
   length() data elements are initialized to <i>initialValue</i>.
 */
template <class T>
inline
RCArray<T>::RCArray( size_t initialSize, T initialVal ) :
  value( new SharedData(initialSize) )
{
  init( initialVal );
}


/**
   Add a new element to the end of the RCArray, resulting
   in length() + 1 data elements.
 */
template <class T>
inline
void RCArray<T>::append( T val )
{
  value->data.append( val );
}

/**
   Return the number of data elements in the RCArray.
 */
template <class T>
inline
size_t RCArray<T>::length() const
{
  return value->data.length();
}


/**
   Return the reference count.
 */
template <class T>
inline
size_t RCArray<T>::getRefCnt()
{
  return value->getRefCnt();
}



template <class T>
inline
T RCArray<T>::operator[](const int ix) const
{
  return value->data[ix];
}



template <class T>
inline
RCArray<T>::ValRef RCArray<T>::operator[](const int ix)
  throw(std::out_of_range)
{
  if (ix < 0 || ix >= value->data.length()) {
    const char *errMsg = "operator[] - index out of bounds";
    throw std::out_of_range( errMsg );
  }
  return ValRef( *this, ix );
}


template <class T>
inline
T RCArray<T>::operator[](const size_t ix) const
{
  return value->data[ix];
}

template <class T>
inline
RCArray<T>::ValRef RCArray<T>::operator[](const size_t ix)
  throw(std::out_of_range)
{
  if (ix >= value->data.length()) {
    const char *errMsg = "operator[] - index out of bounds";
    throw std::out_of_range( errMsg );
  }
  return ValRef( *this, ix );
}


/**
   Return the element at index <i>i</i>
 */
template <class T>
inline
T RCArray<T>::read(const int i) const 
{ 
  return value->data[i]; 
}


/**
   Write value <i>v</i> at index <i>i</i>
 */
template <class T>
inline
void RCArray<T>::write(const int i, const T v) 
{ 
  if (value->isShared()) {
    value = new SharedData( value->data );
  }
  value->data[i] = v; 
}

#endif
