
#ifndef RCBASE_H
#define RCBASE_h

/** \file

   The ValRef object is losely modeled after the Cref class in
   <i>The C++ Programming Language, Third Edition</i> by Bjarne 
   Stroustrup, Section 11.12.  As a result, this code is in the
   public domain.

   \author Ian Kaplan, www.bearcave.com
 */


/**
   An abstract base class for reference counted arrays.  These array
   objects implement copy-on-write semantics.

   This abstract class defines two pure virtual functions, <i>read</i>
   and <i>write</i> and the local ValRef class.  The ValRef class is
   returned for referenced array elements.  This class is a proxy
   which avoids copy-on-read.

   An object that implements copy-on-write arrays defines the
   <i>[]</i> operator (the index operator).  The index operator
   can be used on either the right or left hand size of an
   assignment.  For example, in the statements below <i>a</i> is
   an instance of an object that implements the index operator.

<pre>
       MyObj b = a;      <i>// <b>b</b> is a reference to <b>a</b></i>
       MyType v = a[i];  <i>// A right-hand-side reference</i>
       a[j] = v;         <i>// A left-hand-size reference</i>
</pre>

   References counted objects share a reference to common data
   until an object is modified.  The copy-on-write semantics
   means that a unique copy will be made before shared object
   is modified.  Ideally this saves memory and improves
   performance.  

   In the example above <i>a</i> and <i>b</i> share a reference to the
   same data set.  When <i>a</i> is read, there should be no effect
   (other than the read).  However, when <i>a</i> is modified (by
   being written), a unique copy is made first.  Although previously
   <i>a</i> and <i>b</i> referenced the same data, the write operation
   will force the creation of a unique copy for <i>a</i> and only
   <i>a</i> will be modified.  The <i>b</i> object will retain its
   previous value.

   In a better world there would be some way that the programmer could
   indicate to the C++ compiler that a particular <i>operator []</i>
   function implements the right-hand-side or the left-hand-size
   semantics.  

   The naive implementation (which I used in the first two versions of
   a String class) turns out to be wrong.  Here is the slightly
   simplified code from this somewhat incorrect String class.  My
   intention was that the <i>operator []</i> function below labeled
   "LHS" would implement copy-on-write.  The <i>operator []</i> labeled
   RHS would do a read and nothing more.  

<pre>
      <b>This code does not properly implement copy-on-write</b>

    <i>// 
    // operator [] (RHS) with an integer index
    //</i>
    inline char String::operator [](const int ix) const
    {
       return pShare->str[ix];
    }

     <i>//
     // operator[] (LHS), integer index
     //</i>
     inline char & String::operator [](const int ix)
     {
       makeUnique();  <i>// make a unique copy of the shared data</i>
       return pShare->str[ix];
     } // operator []
</pre>

   As it turned out, the function labeled LHS is called in most
   cases, whether the <i>operator []</i> is on the left or right
   hand side of the expression.  This results in copy-on-read
   which is, obviously, not desirable and destroys much of the utility
   for a reference counted object.

   The proper implementation of the String object uses an instance
   of the ValRef object (instantiated for <tt>char</tt>).

<pre>
     inline
     String::ValRef String::operator[](const int ix)
     {
       return ValRef( *this, ix );
     }
</pre>

<p>
   The ValRef object is losely modeled after the Cref class in
   <i>The C++ Programming Language, Third Edition</i> by Bjarne 
   Stroustrup, Section 11.12.
</p>

   \author Ian Kaplan www.bearcave.com

 */
template <class Type>
class RCBase
{
public:

  class ValRef
  {
    RCBase& o;
    const int i;
  public:
    ValRef(RCBase& oo, const int ii) : o(oo), i(ii) {}
    
    operator Type() const
    {
      // printf("RCBase::Type\n");
      return o.read(i);
    }
    void operator=(Type v)
    {
      // printf("RCBase::operator=\n");
      o.write(i, v);
    }
  }; // class ValRef

  virtual Type read( const int i ) const = 0;
  virtual void write( const int i, const Type v ) = 0;
};

#endif
