
#ifndef RCOBJECT_H
#define RCOBJECT_H

/** \file

   This code is based on Scott Meyers' RCObject class.  This was
   published in Mr. Meyers' book <i>More Effective C++</i> by Scott
   Meyers, Addison-Wesley, 1996.  The RCObject class is part of Scott
   Meyers reference counted String class, published in "Item 29".
   This code is in the public domain.

   \author Ian Kaplan www.bearcave.com

 */

/**
   This the base class for a shared data class that is used in a
   reference counted copy-on-write object (see RCArray).  This class
   encapsulates the reference count.  The shared data class is derived
   from this class and is a container for the the shared data.

   This class is based on the RCObject class published in <i>More
   Effective C++</i> by Scott Meyers, Addison-Wesley, 1996.  This is
   part of Scott Meyers reference counted String class, published in
   "Item 29".  This code is in the public domain.

   I have the first edition and first printing of Mr. Meyers' book.
   I've added some of the changes from Mr. Meyers' errata pages, which
   are not included in this edition.  I've also made some to Mr. Meyers'
   version of RCObject.  In particular, I've removed the shareable flag,
   since as far as I can tell, it is not needed.  I have also made the
   reference count available for testing purposes.

   \author Ian Kaplan www.bearcave.com

 */
class RCObject {
public:
  void addReference();
  void removeReference();
  const bool isShared() const;
  const size_t refCnt() const;

protected:
  RCObject();
  RCObject(const RCObject& rhs);
  RCObject& operator=(const RCObject &rhs);
  virtual ~RCObject();

private:
  size_t refCount;
};

inline
RCObject::RCObject() : refCount(0) {}

inline
RCObject::RCObject(const RCObject& ) : refCount(0) {}

inline
RCObject& RCObject::operator=(const RCObject&)
{
  return *this;
}

inline
RCObject::~RCObject() {}

inline
void RCObject::addReference() { refCount++; }

inline
void RCObject::removeReference()
{ 
  if (--refCount == 0) {
    delete this; 
  }
}

inline
const bool RCObject::isShared() const
{
  return refCount > 1;
}

inline
const size_t RCObject::refCnt() const
{
  return refCount;
}


#endif
