
#ifndef RCPTR_H
#define RCPTR_H

/** \file

    This template was published in <i>More Effective C++</i>
    by Scott Meyers, Addison-Wesley, 1996.

    This is part of Scott Meyers reference counted String class,
    published in "Item 29".  This code is in the public domain.

   \author Ian Kaplan www.bearcave.com
 */

/**
   This template implements a "smart pointer" for reference counted
   objects.  This template is designed to be used with the RCObject
   base class for reference counted objects.

   This template was published in <i>More Effective C++</i>
   by Scott Meyers, Addison-Wesley, 1996.  The RCPtr template
   is part of Scott Meyers reference counted String class,
   published in "Item 29".

   The RCPtr template is used to instantiate a shared data
   object for a reference counted object.  A simplified version
   of the SubString object is shown below.  Note that the
   SubString class variable <i>value</i> is an instance
   of the local shared data object SharedData.

<pre>
     class SubString 
     {
     private:
     
       class SharedData : public RCObject
       {
       private:
         String *pStr_;
         size_t start_;
         size_t subStrLen_;
     
       public:
         SharedData();
       }; // class sharedData
     
       RCPtr&lt;SharedData&gt; value;
     }
</pre>

   The RCPtr "smart pointer" is designed to be instantiated with
   a class that is derived from RCObject.  RCObject support
   the reference count.  The RCPtr "smart pointer" handles assignment
   by deleting the shared data of "this" and assigning it the shared
   data from the right hand side object.  The RCPtr template also
   decrements the reference count when the destructor is called.

   \author Ian Kaplan www.bearcave.com
 */
template <class T>
class RCPtr 
{
public:
  RCPtr(T* realPtr = 0);
  RCPtr(const RCPtr &rhs);
  ~RCPtr();

  bool operator ==(const RCPtr& rhs );
  RCPtr& operator=(const RCPtr& rhs );

  T* operator->() const;
  T& operator*() const;

private:
  T *pointee;
  void init();
};


template <class T>
inline 
void RCPtr<T>::init()
{
  if (pointee != 0) {
    pointee->addReference();
  }
} // init


template<class T>
inline 
RCPtr<T>::RCPtr(T* realPtr) : pointee(realPtr)
{
  init();
}

template<class T>
inline 
RCPtr<T>::RCPtr( const RCPtr& rhs) : pointee( rhs.pointee )
{
  init();
}


template<class T>
inline 
RCPtr<T>::~RCPtr()
{
  if (pointee) {
    pointee->removeReference();
  }
}

/**
   Assign to the "smart pointer".  The "smart pointer" should be
   instantiated with an object derived from RCObject.  The
   RCObject base class supports the removeReference function
   which deallocates "this".  

   In the code below, the new value for the smart pointer is assigned 
   to pointee.  The reference count is incremented to 1 (in init())
   and the old object (oldPointee) is deallocated.
*/
template <class T>
inline 
RCPtr<T>& RCPtr<T>::operator=(const RCPtr &rhs)
{
  if (pointee != rhs.pointee) {
    T *oldPointee = pointee;
    pointee = rhs.pointee;
    init();
    if (oldPointee) {
      oldPointee->removeReference();
    }
  }
  return *this;
}


template <class T>
inline
bool RCPtr<T>::operator ==(const RCPtr &rhs)
{
  bool rslt = (pointee == rhs.pointee);
  return rslt;
}


template <class T>
inline 
T* RCPtr<T>::operator->() const
{
  return pointee;
}


template <class T>
inline 
T& RCPtr<T>::operator*() const
{
  return *pointee;
}


#endif
