
/** \file

   <b>Copyright and Use</b>

   You may use this source code without limitation and without
   fee as long as you include:

<pre>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2001, 2002, 2003
</pre>

   This software is provided "as is", without any warrenty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.

   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>

 */

#include <strstream>
#include <iostream>

using namespace std;

/** POSIX string operations */
#include <string.h>

#include "String.h"


/**
   Initialize the underlying data structure for a String container
   from a C string.
 */
void String::init( const char *initCstr )
{
  if (initCstr != 0) {
    char ch;
    for (ch = *initCstr; ch != '\0'; ch = *initCstr) {
      value->data.append( ch );
      initCstr++;
    }
    value->data.append( ch );
  }
}



/**
   Append a C string at the end of the character string stored in the
   String object.

   The character data stored in a String object should be terminated
   by a null character ('\0').  The null terminator for the "this"
   string will be overwritten by the first character of CStr.

   Reference count and allocation issues should be taken care
   of by the code that calls this function.

 */
void String::appendCString( const char *CStr )
{
  if (CStr != 0) {
    size_t len = length();
    if (len > 0) {
      // if there is a null at the end of the string
      // (which there should be) write over it.
      if (value->data[len-1] == '\0') {
        value->data[len-1] = *CStr++;
      }
    }
    while (*CStr) {
      value->data.append( *CStr++ );
    }
    value->data.append('\0');
  }
} // appendCString


/**
   Append the characters from "s" onto the end of the characters
   stored in the String object.

   Note that when String 's' is appended to the String object
   the null that terminates 's' should be appended as well.

   Reference count and allocation issues should be taken care
   of by the code that calls this function.

 */
void String::appendString( String &s )
{
  size_t i = 0;
  size_t len = s.length();
  size_t thisLen = length(); // length of "this" String
  
  if (thisLen > 0 && len > 0) {
    // write over the null character
    if (value->data[thisLen-1] == '\0') {
      value->data[thisLen-1] = s.value->data[i];
      i++;
    }
  } // if
  while (i < len) {
    char ch = s.value->data[i];
    value->data.append( ch );
    i++;
  } // while
} // appendString



/**
   Compare two strings for equality
 */
int String::compareTo( String &s )
{
  if (length() == 0 && s.length() == 0) {
    // Both String objects contain no data and so are
    // equal
    return 0;
  }

  if (value == s.value) {
      // The two String objects point to the same shared data
      return 0;
  }

  const char *pThisCStr = *this;
  const char *Cstr = s;

  return Cstrcmp( pThisCStr, Cstr );
} // compareTo


/**
   Set the string length of the String to newSize.  This will be the
   size returned by String::strlen().  So the actual amount of data
   allocated will be newSize+1 to account for the null character,
   assuming that newSize > 0.  If newSize == 0, strlen() will be
   zero and the number of data elements will be zero.

   The private function length() will return new_size + 1, since the
   String data is terminated by a null character.  If the String is
   expanded, the String will be padded with spaces.

   If the reference count of the String object is greater than one,
   a unique copy will be made when it is resized, since other 
   copies should be unaffected.

 */
String &String::resize( const size_t newSize )
{
  if (newSize == 0) {
    newEmptyString();
  }
  else { // newSize > 0
    size_t new_len = newSize + 1;
    size_t old_len = length();

    if (value->isShared()) {
      if (new_len >= old_len) {
	// this will copy the data into a new shared data object
	newSharedData();
	value->data.set_size( new_len );
      }
      else { // new_len < old_len
	RCPtr<SharedData> new_value = new SharedData();
	for (size_t i = 0; i < newSize; i++) {
	  new_value->data.append( value->data[i] );
	}
	new_value->data.append('\0');
	// note: smart pointer assignment
	value = new_value;
      }
    }
    else {
      value->data.set_size( new_len );
    }

    for (size_t i = old_len-1; i < new_len-1; i++) {
      value->data[i] = ' ';
    }
    value->data[ new_len-1 ] = '\0';
  }
  return *this;
} // resize



/**
   Compare two C strings pointed to by const char * pointers.
   Return 1 if *pThisCStr is greater than *Cstr, -1 if 
   *pThisCstr is less than *Cstr and 0 if *pThisCstr is equal
   to *Cstr.

 */
int String::Cstrcmp(const char *pThisCStr, const char *Cstr )
{
  int rslt;

  if (pThisCStr == 0) {
    if (Cstr == 0)
      rslt = 0; // both this and CStr are null
    else {
      if (::strlen( Cstr ) == 0)
	rslt = 0;  // compare to an empty string ("")
      else
	rslt = -1; // this < Cstr
    }
  }
  else { // pThisCStr != 0
    if (Cstr == 0)
      rslt = 1; // this > Cstr
    else // pThis != 0 && Cstr != 0
      rslt = strcmp(pThisCStr, Cstr );
  }

  return rslt;
}  // Cstrcmp



/**
   This operator allows the assignment of a String to a <tt>const char
   *</tt>.  For example, in the code below <i>a</i> is assigned to
   <i>pStr</i> in the code below:

<pre>
      String a("wierd operator");
      const char *pStr;

      pStr = a;
</pre>

   The compiler fills in the cast (e.g., const char *), it
   does not have to be explicit.  That is we don't have to
   write

<pre>
      pStr = (const char *)a;  -- unnecessary cast
</pre>

   A C++ compiler should issue an error if an attempt is made
   to assign a String object to a "char *".  The data pointed
   to by the address returned by this operator should never
   be changed since it belongs to the String object.

 */
String::operator const char *() const
{
  const char *pCstr = "";
  size_t len = value->data.length();

  if (len > 0) {
    if (value->data[len-1] != '\0') {
      // then add a NULL terminator
      value->data.append('\0');
    }
    pCstr = value->data.getData();
  }

  return pCstr;
} // operator const char *


/**
   Make "this" into an empty String.  If "this" is shared,
   allocate a new SharedData object.  If "this" is not
   shared, set the length to zero.
 */
void String::newEmptyString()
{
  if (value->isShared()) {
    value = new SharedData();
  }
  else {
    value->data.set_size(0);
  }
} // newEmptyString



/**
   Assign a character to a String
 */
String &String::operator =(const char ch)
{
  newEmptyString();
  if (ch != '\0') {
    value->data.append( ch );
    value->data.append('\0');
  }
  return *this;
} // operator= (String = char)



/**
   Assign a C-string to a String.

   The String(C-string) constructor could be used for this.
   However, using this default generates code which is more
   inefficient.  The write() function is used here since it
   properly handles the String shared data.

   If a null string is assigned to a String, the result will
   be the same as if an empty String were assigned.  For
   example:

<pre>
     char *pCstr = 0;
     String a = "fubar";
     a = pCstr; <i>// same as if "" were assigned</i>
</pre>
 */
String &String::operator =(const char *Cstr)
{
  newEmptyString();
  if (Cstr != 0) {
    char ch;
    for (ch = *Cstr; ch != '\0'; ch = *Cstr) {
      value->data.append(ch);
      Cstr++;
    }
    value->data.append(ch);
  }
  return *this;
} // operator= (String = C-string)



/**
   Concatenate a C string on to the end of the String object.  For
   example:

<pre>
      String s("abcd ");

      s += "efgh";
</pre>

   will result in "s" containing "abcd efgh".

 */
String &String::operator +=( const char *Cstr )
{
  if (value->isShared()) {
    newSharedData();
  }

  // If there is a null character, delete it so we don't
  // append after the null.
  size_t len = value->data.length();
  if (len > 0 && value->data[len-1] == '\0')
    value->data.remove();

  appendCString( Cstr );

  return *this;
} // operator +=



/**
   Concatenate two String objects.

 */
String &String::operator +=(String &s )
{
  if (s.value->data.length() > 0) {
    if (value->isShared()) {
       newSharedData();
    }

    // If there is a null character, delete it so we don't
    // append after the null.
    size_t len = value->data.length();
    if (len > 0 && value->data[len-1] == '\0') {
      value->data.remove();
    }

    appendString( s );
  }
  return *this;
} // operator +=


/*
 * ======== Global binary operators ========
 *
 */

/**
 * Global operator +:  "abcd" + String(" efgh")
 *

   This operator supports an operation like "abcd" + String(" efgh").
   The result of this operator would be a String object containing
   "abcd efgh".

 */
String operator +(const char *Cstr, String &s)
{
  String tmp;
  if (Cstr != 0) {
    tmp = Cstr;
  }
  
  tmp += s;
  return tmp;
} // global operator +


/**
  Global operator +: Concatenate a character and a String object.

<pre>
     String Str("ab");
     String a;

     a = 'c' + Str
</pre>

   Here, "a" will contain "abc"

 */
String operator +(const char ch, String &s)
{
  String tmp;

  tmp = ch;
  tmp += s;

  return tmp;
} // global operator +


/**
   Global Cstr == String operator, where Cstr is a
   "C" string.
 */
bool operator ==(const char *Cstr, String &s)
{
  return (s == Cstr);
} // global operator ==


/**
   Global Cstr != String operator, where Cstr is a
   "C" string.
 */
bool operator !=(const char *Cstr, String &s)
{
  return (s != Cstr);
} // operator !=


/**
 *   Cstr <= s: however, we must use the
 *              comparision s to Cstr.
 *              So this expression is true
 *              if s >= Cstr.
 */
bool operator <=(const char *Cstr, String &s)
{
  return (s >= Cstr);
} // global operator <=


/**
 *   Cstr >= s: however, we must use the
 *              comparision s to Cstr.
 *              So this expression is true
 *              if s <= Cstr.
 */
bool operator >=(const char *Cstr, String &s)
{
  return (s <= Cstr);
} // global operator >=


/**
 *   Cstr < s: however, we must use the
 *             comparision s to Cstr.
 *             So this expression is true
 *             if s > Cstr.
 */
bool operator <(const char *Cstr, String &s)
{
  return (s > Cstr);
} // global operator <


/**
 *   Cstr > s: however, we must use the
 *             comparision s to Cstr.
 *             So this expression is true
 *             if s < Cstr.
 */
bool operator >(const char *Cstr, String &s)
{
  return (s < Cstr);
} // global operator >


