
#ifndef MY_STRING_H
#define MY_STRING_H
// don't want a collision with another string macro, so I have
// not used STRING_H.

/** \file

   <b>Copyright and Use</b>

   You may use this source code without limitation and without
   fee as long as you include:

<pre>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2001, 2002, 2003
</pre>

   This software is provided "as is", without any warrenty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.

   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>

 */

#include <stdexcept>

#include <stdio.h>

#include "SubString.h"
#include "RCArray.h"

/**
   Define a type for a growable reference counted array object
   that implements copy-on-write semantics.
 */
typedef RCArray<char> CharArray;

/**

   This class is a container for strings composed of 8-bit ASCII
   characters.  This class uses reference counts and copy-on-write
   to improve memory use and, perhaps, performance.

   This is the version 3.0 of this class.  This class was originally
   modeled after the Rogue Wave RWCString class (using the
   documentation available on the Rogue Wave web site).

   <b>Why Yet Another String Container</b>

   Although I like the Rogue Wave class and the RWCString class in
   particular, I don't use the Rogue Wave classes in the software I
   write for Bear Products International.  There are several reasons
   for this.  I did not want to pay the Rogue Wave license fees, which
   are high for a <a href="http://www.bearcave.com">Cave</a> based
   software developer.  Nor did I want to force anyone who licenses my
   software to also Rogue Wave license fees (sorry guys).  I also
   wanted software that was as transparent as possible.  Rogue Wave
   classes tend to be incompletely documented and it is not always
   easy to fully understand the way a particular class functions.

   Rogue Wave is not the only alternative when it comes to reference
   counted String containers.  The Microsoft Foundation Classes
   include a string container and some versions of the STL template
   library implement the STL string container as a reference counted
   object.  Each of these alternatives has some draw back.  Microsoft
   Foundation Class code is not portable outside of Windoz.  Also,
   with the new .NET framework, it seems possible that Microsoft would
   at some point phase out the Foundation Classes.  The reference
   counted versions of the STL string container that I'm aware of also
   incur a license fee, although not as steep as Rogue Wave (the 
   Dinkumware STL library is one example of an STL implementation that
   includes reference counted strings).

   Finally, to quote Bjarne Staroustrup

<p>
     <i>Writing string classes has great educational value...</i><br>
     Chapter 20, <i>The C++ Programming Language</i>, third edition
     Bjarne Staroustrup
</p>

   The String class implements a sub-string (string section) operation
   which results in a SubString object.  As a result, the String
   and SubString classes are closely related.

   This class comes with a set of regression tests.  I recommend that
   you run these regression tests with your compiler, on your system.
   Not all compilers implement C++ in the same way.

   \author Ian Kaplan www.bearcave.com

*/
class String : public CharArray
{
  // Make SubString a "friend" so it can access the protected read() and
  // write functions.
  friend SubString;
protected:
  void init(const char *initCstr );
  size_t length();
  void appendCString( const char *CStr );
  void appendString( String &Str );
  void newEmptyString();
  void newSharedData();
  int Cstrcmp( const char *s1, const char *s2 );
  int compareTo( String &s );

public:

  String();
  String( const char ch );
  String( const char *initCstr );
  String( const String &Str );

  // sub-string operator (can be used on both lhs and rhs)
  SubString operator()(const size_t start, const size_t len);

  /** strlen does not include the null at the end of the string */
  size_t strlen();

  size_t getRefCnt();

  int compareTo( const char *Cstr );

  String &resize(const size_t new_size );

  //
  // operators
  //
  operator const char*() const;

  String &operator =(const char ch);        // String = char
  String &operator =(const char *Cstr);     // String = C-string
  String &operator =(const SubString &sub); // String = SubString

  String operator +(const char *Cstr ); // concatenate a String and a C string
  String operator +(String &s );        // concatenate two Strings
  String operator +(const char ch );    // concatenate a String and a char
  // concatenate a String and a SubString
  String operator +(const SubString &sub); 

  String &operator +=(const char *Cstr ); // concatenate a C string to a String
  String &operator +=(String &s );        // concatenate a String to a String
  String &operator +=(const char ch);     // concatenate a char to a String

  bool operator ==(const char *Cstr ) { return (compareTo( Cstr ) == 0); }
  bool operator ==(String &s )  { return (compareTo( s ) == 0); }
  bool operator !=(const char *Cstr ) { return (compareTo( Cstr ) != 0); }
  bool operator !=(String &s )  { return (compareTo( s ) != 0); }
  bool operator <=(const char *Cstr ) { return (compareTo( Cstr ) <= 0); }
  bool operator <=(String &s )  { return (compareTo( s ) <= 0); }
  bool operator >=(const char *Cstr ) { return (compareTo( Cstr ) >= 0); }
  bool operator >=(String &s )  { return (compareTo( s ) >= 0); }
  bool operator <(const char *Cstr )  { return (compareTo( Cstr ) < 0); }
  bool operator <(String &s )   { return (compareTo( s ) < 0); }
  bool operator >(const char *Cstr )  { return (compareTo( Cstr ) > 0); }
  bool operator >(String &s )   { return (compareTo( s ) > 0); }

}; // String

/*
 * Global operators
 *
 */
String operator +(const char *Cstr, String &s);
String operator +(const char ch, String &s);

bool operator ==(const char *Cstr, String &s);
bool operator !=(const char *Cstr, String &s);
bool operator <=(const char *Cstr, String &s);
bool operator >=(const char *Cstr, String &s);
bool operator <(const char *Cstr, String &s);
bool operator >(const char *Cstr, String &s);

inline
String::String() : CharArray() {}

/** String constructor with a character argument */
inline
String::String( const char ch ) : CharArray()
{ 
  if (ch != '\0') {
    value->data.append( ch );
    value->data.append('\0');
  }
}

/** String constructor with a C-string argument */
inline
String::String( const char *initCstr ) : CharArray()
{
  init( initCstr );
}

/**
   String copy constructor
 */
inline
String::String( const String &Str ) : CharArray()
{
  value = Str.value;
}


/**
   Return the number of data elements (characters) consumed
   by the String.  This includes the null character.
   So length() == strlen() + 1
 */
inline
size_t String::length()
{
  return CharArray::length();
}


/**
   Return the reference count.  This useful primarily for debug
   and verification.  In the perfect world of bug free software
   the reference count could be entirely hidden.
 */
inline
size_t String::getRefCnt()
{
  return CharArray::getRefCnt();
}




/**
 Sub-string operator: this function supports the left hand side
 (l-value) form for the sub-string operation.  For example:

<pre>
      String a("abcdefghijkl");

      a(4, 5) = "12345";
</pre>

   The String object "a" now contains "abcd12345jkl"

   This function returns a SubString object.  The SubString object
   supports the assignment operator, which will assign the
   C-string to <i>a</i>

 */
inline
SubString String::operator ()(const size_t start, const size_t len)
{
  SubString subStr( *this, start, len );

  return subStr;
}


/**
   Append a String and a SubString

<pre>
   String a = "abcde";
   // indices: 01234567
   String b = "123 5678";
   String c = a + b(3,4); // result will be "abcde 567"
</pre>

 */
inline
String String::operator +(const SubString &sub)
{
  String section = sub;
  String result = *this + section;
  return result;
} // operator +



/**
   Assign a SubString to a String.  For example:
<pre>
     String a = "lazy dog";
     String b = a(5, 3);
</pre>
 */
inline
String &String::operator =(const SubString &sub)
{
  *this = (String)sub;
  return *this;
} // operator =


/**
   Allocate a new SharedData object which contains a copy of the old
   data.  Note that this code relies on the fact that <i>value</i> is
   a "smart pointer" which will decrement the reference count and
   deallocate the space if the reference count is zero.  
*/
inline
void String::newSharedData()
{
  value = new SharedData( value->data );
}



/**
   Compare the character string in "this" object with the
   argument Cstr.
<pre>
     if (this < Cstr)  return -1;
     if (this == Cstr) return  0;
     if (this > Cstr)  return  1;
</pre>

 */
inline 
int String::compareTo( const char *Cstr )
{
  const char *pThisCStr = *this;
  
  return Cstrcmp( pThisCStr, Cstr );
}



/**
  Append a character to the end of "this" string

 */
inline 
String &String::operator +=(const char ch)
{
  char buf[4];

  buf[0] = ch;
  buf[1] = '\0';

  *this += buf;   // call String += Cstr
  return *this;
} // operator +=



/**
 * operator +: String("abcd") + " efgh"
 *

   Create a new String object which contains the concatenation of the
   two operand strings (self and the argument "s").  The reference
   count for the new String object will be zero.  If it is assigned
   then it will be incremented.

   Example:

<pre>
      String a("abcd");
      String b = a + " efgh";
</pre>

   When these statements are executed String 'a' will contain "abcd"
   and String 'b' will contain "abcd efgh".  The reference count
   for both String objects will be 1.

 */
inline 
String String::operator +( const char *Cstr )
{
  String tmp;
  
  tmp.appendString( *this );
  tmp.appendCString( Cstr );

  return tmp;
} // operator +


/**
 * operator +: String("abcd") + String(" efgh")
 *
 */
inline 
String String::operator +(String &s)
{
  String tmp;
  
  tmp.newSharedData();

  tmp.appendString( *this );
  tmp.appendString( s );
  
  return tmp;
} // operator +


/**
 * operator+ : String + ch
 *

   Append a String and a character.   

 */
inline 
String String::operator +(const char ch)
{
  char buf[4];

  buf[0] = ch;
  buf[1] = '\0';

  String tmp;

  tmp = *this + buf;  // call String + Cstr

  return tmp;
} // operator +



/**
   strlen() returns the number of characters in the string,
   not including the null character which terminates the
   String.

 */
inline
size_t String::strlen()
{
  int len = value->data.length(); 
  if (len > 0 && value->data[len-1] == '\0')
    len--;
  return len;
}


#endif
