
/** \file
   Regression tests for the SubString class.

   Note that the existence of these regression tests does not change
   that statement <i>this code is provided without waranty</i>.  You
   use this code at your own risk.  

    Compiling this code:

    This code is written to compile on Microsoft Visual C++ 6.0 and
    higher.  It will also compile on GNU C++ and Sun's 6.0 and later
    C++ compilers.  To compile under Microsoft VC++:

<pre>
      cl -Zi -TP -GX SubString.C String.C SubStrTest.C -o SubStrTest
</pre>

    The -GX flag is needed to support exceptions, which are used 
    rather than the assert() functions in the earlier versions.

   Copyright and Use

   You may use this source code without limitation and without
   fee as long as you include:

<pre>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2001.
</pre>

   This software is provided "as is", without any warrenty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.

   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>


 */


#include <stdio.h>

#include "SubString.h"


bool SubStringPassByValue( SubString subStr, size_t previousRefCnt )
{
  bool result = (subStr.getRefCnt() == previousRefCnt + 1);
  if (result == false) {
    printf("SubStringPassByValue: argument subStr.getRefCnt() = %d\n",
	   subStr.getRefCnt() );
    printf("SubStringPassByValue: previousRefCnt = %d\n", previousRefCnt );
  }
  return result;
}


bool SubStringPassByReference( SubString &subStr, size_t previousRefCnt )
{
  bool result = (subStr.getRefCnt() == previousRefCnt);
  return result;
}


/**
   Test to verify that reference counting is handled properly for
   the SubString class.  This also tests assignment as a side-effect.
 */
bool subStrRefCnt()
{
  bool result = true;
  printf("Test SubString reference count\n");

  String a("abcdefghijk");

  //
  // Test reference count
  //
  SubString subStrA(a, 5, 6);
  if (subStrA.getRefCnt() != 1) {
    printf("1. reference count is %d, should be 1\n",
	   subStrA.getRefCnt());
    result = false;
  }

  size_t previousRefCnt = subStrA.getRefCnt();
  if (! SubStringPassByReference( subStrA, previousRefCnt) ) {
    printf("2. Pass by reference failed\n");
    result = false;
  }

  previousRefCnt = subStrA.getRefCnt();
  if (! SubStringPassByValue( subStrA, previousRefCnt) ) {
    printf("3. Pass by reference value\n");
    result = false;
  }

  if (a.getRefCnt() != 1) {
    printf("4. refCnt of the String referenced by the SubString = %d, should be 1\n", a.getRefCnt() );
    result = false;
  }

  return result;
} // subStrRefCnt



/**
   Test SubString section operations

   The SubString constructor creates a section:

<pre>
       String s = "Mr. Golden Noodle";
       String pasta = s(11, 6);
</pre>

   were 11 is the start index and 6 is the length.

   The construction of a SubString should always result in a correct
   SubString.  This means that the start index should be within the
   bounds of the String and the start index plus the size should be in
   the bounds as well.  If this is not the case, the SubString 
   constructor will throw an exception.

   So, what about:

<pre>
     String empty;
     String empty2 = empty(0, 0);
</pre>

   Ok, so we're taking a zero length section of a String starting at
   zero.  Is that so wrong?  I don't know about <i>so</i> wrong, but
   it does seem incorrect.  The rules with with a SubString is that
   the starting index must be less than the String length.  In this
   case the starting index is equal to the String length.

*/
bool subStrSection()
{
  printf("Test SubString section operations\n");
  bool rslt = true;
  
  String a("abcdefghij");
  String d = a;

  // a(0,5) means start at index 0, length of the section is 5
  String subA = a(0,5);
  if (subA.strlen() != 5) {
    printf("1. SubString length = %d, should be 5\n", subA.strlen() );
    rslt = false;
  }

  // make sure that the reference count for the original string
  // is not affected.  Note that a right hand size SubString 
  // operation does not affect the String on which the SubString
  // is defined.
  if (d.getRefCnt() != 2) {
    printf("2. d.getRefCnt() = %d, should be 2\n", d.getRefCnt() );
  }
  
  String b = subA;
  if (b != "abcde") {
    printf("3. SubString section is incorrect. ");
    const char *Cstr = b;
    printf("SubStr = %s, should be \"abcde\"\n", Cstr );
    rslt = false;
  }

  String c = a(5,5);
  if (c != "fghij") {
    printf("4. SubString section is incorrect. ");
    const char *Cstr = b;
    printf("SubStr = %s, should be \"abcde\"\n", Cstr );
    rslt = false;
  }

  if (a.getRefCnt() != 2) {
    printf("5. a.getRefCnt() = %d, should be 2\n", a.getRefCnt() );
    rslt = false;
  }

  bool exceptionTest1 = false;
  // Test exceptions on an out of bound reference
  try {
    // use a length (8) that places the section beyond the end
    // of the String.
    String x = a(5, 8);
  }
  catch (std::out_of_range e) {
    exceptionTest1 = true;
  }

  if (!exceptionTest1) {
    printf("6. There should have been a SubString construction exception\n");
    rslt = false;
  }

  bool exceptionTest2 = false;
  try {
    // use a start that is beyond the end of the String
    String x = a(10, 1);
  }
  catch (std::out_of_range e) {
    exceptionTest2 = true;
  }

  if (!exceptionTest2) {
    printf("7. There should have been a SubString construction exception\n");
    rslt = false;
  }

  String empty;
  if (empty.strlen() != 0) {
    printf("8. length of an empty string is wrong\n");
    rslt = false;
  }

  bool exceptionTest3 = false;
  String empty2;
  try {
    empty2 = empty(0,0);
  }
  catch(std::out_of_range e) {
    exceptionTest3 = true;
  }

  if (!exceptionTest3) {
    printf("9. There should have been a SubString construction exception\n");
    rslt = false;
  }

  if (empty2.strlen() != 0) {
    printf("10. length of empty2 is wrong\n");
    rslt = false;
  }

  if (empty2.getRefCnt() != 1) {
    printf("11. reference count of empty2 is wrong\n");
    rslt = false;
  }

  String e = "12345678";
  String f = e;
  e(4,0) = "abcd";
  if (e != "1234abcd5678") {
    printf("12. insert, via a zero length section, failed\n");
    size_t len = e.strlen();
    printf("12. e.strlen() = %d\n", len );
    if (len > 0) {
      printf("12. e = %s\n", (const char *)e );
    }
  }

  if (e.strlen() != 12) {
    printf("13. e.strlen() = %d, should be 12\n", e.strlen() );
  }

  if (f != "12345678") {
    printf("14. improperly altered string in 'f'\n");
  }

  if (f.getRefCnt() != 1 && e.getRefCnt() != 1) {
    printf("15. reference counts are wrong\n");
  }
  
  return rslt;
} // subStrSection



/**
   Test assignment to and from a SubString object
 */
bool subStrAssign()
{
  bool rslt = true;

  printf("Test assignment to and from a SubString object\n");
  const char *CStr1 = "abcde1234jklm";
  const char *CStr2 = "abcdefghijklm";
  String a = CStr1;

  if (a != CStr1) {
    printf("1. Contents of String is incorrect\n");
    rslt = false;
  }

  String b = a;
  if (b.getRefCnt() != 2) {
    printf("2. Reference count is incorrect\n");
    rslt = false;
  }

  a(5, 4) = "fghi";
  if (a != CStr2) {
    printf("3. Contents of String is incorrect\n");
    rslt = false;
  }

  if (b.getRefCnt() != 1) {
    printf("4. Reference count is incorrect\n");
    rslt = false;
  }
  
  String c = a(0, 5);
  if (c != "abcde") {
    printf("5. Contents of String is incorrect\n");
    rslt = false;
  }

  // assign a string to a substring section
  a(5, 5) = c;
  if (a(0,10) != "abcdeabcde") {
    printf("6. Contents of String is incorrect\n");
    rslt = false;
  }

  // Now check assignment to regions that are not the same
  // length as the string being assigned.

  String target = "the lazy golden";
  String t2 = target;

  // assign a longer C-string to a shorter assignment region
  t2(0, 3) = "morgie is a";
  if (t2 != "morgie is a lazy golden") {
    printf("7. C-string assignment to SubString is wrong\n");
    rslt = false;
  }

  if (target.getRefCnt() != 1) {
    printf("8. Reference count is wrong\n");
    rslt = false;
  }

  if (target != "the lazy golden") {
    printf("9. The contents of target is wrong\n");
    rslt = false;
  }

  t2(17, 6) = "spaniel";
  // assign a shorter C-string to a longer assignment region
  t2(0, 6) = "max";
  if (t2 != "max is a lazy spaniel") {
    const char *Cstr = t2;
    printf("10. The contents of t2 is incorrect.  t2 = %s\n", Cstr );
    rslt = false;
  }

  // Test assignments of String to SubString

  // Longer assigned to shorter
  target = "max is a dog";
  String replaceAnimal = "people";
  String replaceName = "Maggie";
  target(9,3) = replaceAnimal;
  target(0,3) = replaceName;
  if (target != "Maggie is a people") {
    printf("11. String to SubString assignment failed\n");
    rslt = false;
  }

  // shorter assignments to longer SubString region
  replaceName = "Ian";
  target(0, 6) = replaceName;

  if (target != "Ian is a people") {
    printf("11. String to SubString assignment failed\n");
    rslt = false;
  }

  if (target.strlen() != strlen("Ian is a people")) {
    printf("12. length of the string is wrong\n");
    rslt = false;
  }

  // Test the explicit SubString constructor with assignment.
  // Also make sure that chained assignment works in this case.

  const char *expected_result = "the lazy golden";
  String u;
  String s = "the lazy dog";
  SubString t(s, 9, 3);

  u = t = "golden";
  
  if (s != expected_result) {
    printf("13. the contents of s is \"%s\", should be \"%s\"\n",
	   (const char *)s, expected_result);
    rslt = false;
  }

  if (u != expected_result) {
    printf("14. the contents of u is \"%s\", should be \"%s\"\n",
	   (const char *)u, expected_result);
    rslt = false;
  }

  // The result of the assignment to the SubString should be the
  // String, "s" passed to the constructor.  This is assigned to
  // the String u, so the reference count should be 2.
  if (u.getRefCnt() != 2) {
    printf("15. reference count is %d, should be 2\n", u.getRefCnt() );
    rslt = false;
  }

  return rslt;
} // subStrAssign


/**
   Test SubString relational operations
 */
bool subStrRelations()
{
  bool rslt = true;
  String a = "abcdeabcde";
  printf("Test SubString relational operators\n");

  //
  //  == operator
  //

  // SubString != SubString
  bool t1 = false;
  if (a(0,5) == a(5, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("1. SubString == SubString failed\n");
    rslt = false;
  }

  // SubString == Cstr 
  bool t2 = false;
  if (a(0,5) == "abcde") {
    t2 = true;
  }
  if (!t2) {
    printf("2. SubString == Cstr failed\n");
    rslt = false;
  }

  // SubString == String
  bool t3 = false;
  String b = "abcde";
  if (a(0,5) == b) {
    t3 = true;
  }
  if (!t3) {
    printf("3. SubString == String failed\n");
    rslt = false;
  }

  t1 = true;
  if (a(1,4) == "bcd") {
    t1 = false;
  }
  if (!t1) {
    printf("4. SubString == CStr of unequal length should have been false\n");
    rslt = false;
  }


  //
  // != operator
  //
  t1 = false; 
  t2 = false; 
  t3 = false;
  //          012345678901234
  String c = "abcdexyzzyabcde";
  // SubString != SubString
  if (a(0,5) != c(5,5)) {
    t1 = true;
  }
  if (!t1) {
    printf("5. SubString != SubString failed\n");
    rslt = false;
  }

  // SubString != Cstr
  if (a(0,5) != "xyzzy") {
    t2 = true;
  }
  if (!t2) {
    printf("6. SubString != SubString failed\n");
    rslt = false;
  }
  // SubString != String
  b = "zyzzy";
  if (a(0,5) != b) {
    t3 = true;
  }
  if (!t3) {
    printf("7. SubString != String failed\n");
    rslt = false;
  }

  bool t4 = false;
  if (a(0, 5) != "abcd") {
    t4 = true;
  }
  if (!t4) {
    printf("8. SubString != CStr of unequal length failed\n");    
    rslt = false;
  }

  t4 = true;
  if (a(1, 4) != "bcde") {
    t4 = false;
  }
  if (!t4) {
    printf("9. SubString != CStr are equal - test should be false\n");
    rslt = false;
  }

  //
  // < operator
  //
  t1 = false;
  t2 = false;
  t3 = false;
  t4 = false;

  // SubString < SubString
  if (a(1,5) < c(5,5)) {
    t1 = true;
  }
  if (!t1) {
    printf("10. SubString < SubString failed\n");
    rslt = false;
  }

  // SubString < SubString: overlapping SubString sections on the same String
  t1 = false;
  if (c(1,5) < c(5,5)) {
    t1 = true;
  }
  if (!t1) {
    printf("11. SubString < SubString failed for overlapping sections\n");
    rslt = false;
  }

  // SubString < Cstr
  t1 = false;
  if (c(1,5) < "pdsrsmzf") {
    t1 = true;
  }
  if (!t1) {
    printf("12. SubString < Cstr \n");
    rslt = false;
  }

  // SubString < String
  // note that b is still "xyzzy"
  t1 = false;
  if (c(0,8) < b) {
    t1 = true;
  }
  if (!t1) {
    printf("13. SubString < String \n");
    rslt = false;
  }

  //
  // > operator
  //

  // SubString > SubString
  //   0123456789
  String d = "    abc   ";
  t1 = false;
  if (c(10, 5) > d(2, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("14. SubString > SubString\n");
    rslt = false;
  }
  
  // SubString > Cstr (note, "xyz" will be "extended" with null chars)
  t1 = false;
  if (c(5, 5) > "xyz") {
    t1 = true;
  }
  if (!t1) {
    printf("15. SubString > Cstr\n");
    rslt = false;
  }

  // SubString > String
  b = "lesser of two evils";
  t1 = false;
  if (c(5, 5) > b) {
    t1 = true;
  }
  if (!t1) {
    printf("16. SubString > String\n");
    String ts = c(0, 5);
    const char *t1 = ts;
    const char *t2 = b;
    printf("c(0, 5) = %s, b = %s\n", t1, t2);
    rslt = false;
  }

  //
  // >= operator
  //
  t1 = false;
  if (c(0,5) >= c(10, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("17. SubString >= SubString\n");
    rslt = false;
  }

  t1 = false;
  if (c(0,5) >= "abcde") {
    t1 = true;
  }
  if (!t1) {
    printf("18. SubString >= CStr\n");
    rslt = false;
  }

  t1 = false;
  if (c(5,5) >= "abcde") {
    t1 = true;
  }
  if (!t1) {
    printf("19. SubString >= CStr\n");
    rslt = false;
  }


  t1 = false;
  b = "abcde";
  if (c(0,5) >= b) {
    t1 = true;
  }
  if (!t1) {
    printf("20. SubString >= String\n");
    rslt = false;
  }

  //
  // <= operator
  //
  t1 = false;
  if (c(0,5) <= c(5, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("21. SubString <= SubString\n");
    rslt = false;
  }

  t1 = false;
  if (c(0,5) <= "xyzzy") {
    t1 = true;
  }
  if (!t1) {
    printf("22. SubString <= CStr\n");
    rslt = false;
  }

  t1 = false;
  b = "xyzzy";
  if (c(0,5) <= b) {
    t1 = true;
  }
  if (!t1) {
    printf("23. SubString >= String\n");
    rslt = false;
  }

  t1 = false;
  b = "abcde";
  if (c(0,5) <= b) {
    t1 = true;
  }
  if (!t1) {
    printf("24. SubString >= String\n");
    rslt = false;
  }

  return rslt;
} // subStrRelations



/**
   Test for other types on the LHS of a relation.  This 
   must be supported by global operators.
 */
bool globalSubStrRelations()
{
  printf("Test global SubString relational operators\n");
  bool rslt = true;

  //          012345678901234
  String a = "abcdexyzzyabcde";

  // Cstr == SubString
  bool t1 = false;
  if ("abcde" == a(0, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("1. Cstr == SubString\n");
    rslt = false;
  }

  // Cstr != SubString
  t1 = false;
  if ("abcde" != a(5, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("2. Cstr != SubString\n");
    rslt = false;
  }

  // Cstr < SubString
  t1 = false;
  if ("abcde" < a(5, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("3. Cstr < SubString\n");
    rslt = false;
  }

  // Cstr > SubString
  t1 = false;
  if ("xyzzy" > a(0, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("4. Cstr > SubString\n");
    rslt = false;
  }

  // Cstr >= SubString
  t1 = false;
  if ("abcde" >= a(0, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("5. Cstr >= SubString\n");
    rslt = false;
  }

  // Cstr >= SubString
  t1 = false;
  if ("xyzzy" >= a(0, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("6. Cstr >= SubString\n");
    rslt = false;
  }

  // Cstr <= SubString
  t1 = false;
  if ("abcde" <= a(0, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("7. Cstr <= SubString\n");
    rslt = false;
  }

  // Cstr <= SubString
  t1 = false;
  if ("abcde" <= a(5, 5)) {
    t1 = true;
  }
  if (!t1) {
    printf("8. Cstr <= SubString\n");
    rslt = false;
  }

  return rslt;
}  // globalSubStrRelations


int
main()
{
  bool passed = true;
  printf("SubString tests \n");

  if (!subStrRefCnt()) {
    passed = false;
  }

  if (!subStrSection()) {
    passed = false;
  }

  if (!subStrAssign()) {
    passed = false;
  }

  if (!subStrRelations()) {
    passed = false;
  }

  if (! globalSubStrRelations()) {
    passed = false;
  }

  printf("Tests ");
  if (passed)
    printf("passed\n");
  else
    printf("failed\n");
  return 0;
}
