/** \file

   <b>Copyright and Use</b>

   You may use this source code without limitation and without
   fee as long as you include:

<pre>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2001, 2002, 2003.
</pre>

   This software is provided "as is", without any warrenty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.

   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>

 */

/** POSIX string operations */
#include <string.h>

#include "String.h"
#include "SubString.h"

/**
   User defined type conversion between SubString and String.
   For example

<pre>
       String a("abcdefg");
       String b = a(0,4);
</pre>

   The string() function sections the SubString, resulting in a String.

 */
SubString::operator String() const
{ 
  return this->string(); 
}


/**
   Check that start and len are within the bounds of the String
   argument <i>s</i>.
 */
void SubString::rangeCheck(String &s, const size_t start, const size_t len)
     throw(std::out_of_range)
{
  char *errMsg = 0;
  bool rangeError = false;
  size_t sLen = s.strlen();
  if (start >= sLen) {
    errMsg = "SubString: start >= String length";
    rangeError = true;
  }
  else if (start + len > sLen) {
    errMsg = "SubString: start + len >= String length";
    rangeError = true;
  }

  if (rangeError) {
    throw std::out_of_range( errMsg );
  }
} // rangeCheck


/**
   This constructor initializes the SubString object with a
   reference to a String object, a start index for the SubString
   section and a length.  For example:

<pre>
      String s = "max is a cat";
      SubString t(s, 9, 3);
</pre>

   Except for initialization, nothing happens in the SubString
   constructor.  It is only when a SubString is assigned or otherwise
   operated on that a new String is created (the region of the
   initialization String defined by the start and length).

   SubStrings are commonly created by the String class.  When creating
   a SubString object the String class initializes the SubString with
   itself, and with the starting index and length.  The code below
   shows an example of an sub-string expression created by the String
   class.

<pre>
       String a("abcd12345jkl");

       String b = a(4, 5);
</pre>

   The first argument is the starting index (where Strings start
   with index 0).  The second argument is the length.

   This class throws the std::out_of_range error when either the start
   or (start+len) are out of range (beyond the end of the String).
   Exceptions are described in <i>The C++ Programming Language, Third
   Edition</i> by Stroustrup, section 14.10.

   One of the corner cases that must be considered is a SubString 
   of an empty String:

<pre>
     String empty; <i>// empty string </i>
     String empty2 = empty(0, 0); <i>// This will throw an exception</i>
</pre>

   Ok, so we're taking a zero length section of a String starting at
   zero.  Is that so wrong?  I don't know about <i>so</i> wrong, but
   it does seem incorrect.  The rules with with a SubString is that
   the starting index must be less than the String length.  In this
   case the starting index is equal to the String length.  As a result
   this example would throw an exception.

   A SubString can be thought of as copy on demand.  The SubString
   expression will not actually section the String until the
   SubString is assigned, either explicitly or implicitly.

   The SubString will not alter the reference count of the String
   that it references.
      
 */
SubString::SubString( String &s, const size_t start, const size_t len )
          throw(std::out_of_range)
          : value( new SharedData() )
{
  try {
    rangeCheck(s, start, len);
    size_t sLen = s.strlen();
    value->pStr( &s );
    value->start( start );
    value->subStrLen(len);
  }
  catch (std::out_of_range e) {
    throw e;
  }
} // SubString constructor



/**
   A SubString defines a region in a String.  When an assignment is
   made to a SubString and the length of the data being assigned is
   not the same as the SubString length, a copy space must be
   created.

   In the case where the length of the data being copied is larger
   than the SubString length, the area defined by the SubString must
   be made larger.  This extends the length of the string and moves
   characters above the region to higher indices.

   In the case where the length of the data being copied is less
   than the SubString length, the area defined by the SubString
   must be reduced.  This reduces the length of the String and
   results in moving characters above the copy region to lower
   indices.

   This function allows assignment code to simply copy data into
   the assignment region.  Note that both the write() and resize()
   functions will create an unshared String if the String is shared.

*/
void SubString::createCopySpace(size_t start, size_t charsToCopy)
{
    size_t subStrLen = value->subStrLen();

    if (charsToCopy != subStrLen) {
      size_t strlen = value->pStr()->strlen();
      if (charsToCopy < subStrLen) {
	// The end of the copy region will be start+charsToCopy
	// The end of the SubString section is start+subStrLen
	size_t dest = start+charsToCopy;
	for (size_t i = start+subStrLen; i < strlen; i++, dest++) {
	  char ch = value->pStr()->read(i);
	  value->pStr()->write(dest, ch );
	}
	size_t shrink = subStrLen - charsToCopy;
	size_t newLen = strlen - shrink;
	value->pStr()->resize( newLen );
      }
      else if (charsToCopy > subStrLen) {
	size_t expand = charsToCopy - subStrLen;
	size_t newLen = strlen + expand;
	value->pStr()->resize( newLen );

	size_t dest = newLen-1;
	for (int i = strlen-1; i >= (int)start; i--, dest--) {
	  char ch = value->pStr()->read(i);
	  value->pStr()->write(dest, ch );
	}
      }
    }
} // createCopySpace

/**
 String = SubString + Cstr
*/
String SubString::operator +(const char *Cstr)
{
  String result = *this;
  if (Cstr != 0) {
    result += Cstr;
  }
  return result;
}


/**
   String = SubString + SubString
*/
String SubString::operator +(const SubString& sub)
{
  String s1 = *this;
  String s2 = sub;
  String result = s1 + s2;
  return result;
}


/**
   String = SubString + String
 */
String SubString::operator +(const String &str)
{
  String s1 = *this;
  String result = s1 + str;
  return result;
}


/**
   operator =: This function assigns a C character string to a
   SubString.  For example:

<pre>
      String foo("abcdefghijkl");
      String bar = foo;

      foo(4, 4) = "1234";
</pre>

   After the SubString assignment is executed, the String "foo" will
   contain "abcd1234ijkl".  Note that the contents of <i>bar</i> will
   be unchanged.  However, the reference count of <i>bar</i> will be
   1.

   A smaller string can be assigned to a larger string:

<pre>
      foo(4, 4) = "12";
</pre>

   will result in "abcd12ijkl" (note that in this case a length
   4 string ("efgh") was replaced by a length 2 string ("12"))

   The symetric case, where a longer string is assigned to a 
   smaller string:

<pre>
      foo(4, 4) = "123456
</pre>

   Results in a longer String, as the length 4 string is replaced by
   a length 6 string: "abcd123456ijkl".

   Assignment to a SubString results in a String.  The String is the
   String that was used to initialize the SubString constructor,
   modified by the assignment.  The assignment

<pre>
      foo(4, 4) = "1234";
</pre>
   
   can be broken up into two steps:

<ul>
     <li>
     SubString creation: 
     <pre>
         SubString t1(foo, 4, 4);
     </pre>
     </li>
     <li>
     Assignment to the SubString, resulting in a string:
     <pre>
         String t2 = (t1 = "1234");
     </pre>
     </li>
</ul>

   There are three cases that are dealt with in this assignment:
<ul>
  <li>
  The length of the Cstr is equal to the length of the SubString
  section.  In this case the Cstr can simply be copied into the
  String.
  </li>
  <li>
  The length of the Cstr is less than the length of the SubString
  section.  In this case the String is collapsed before the assignment.
  </li>
  <li>
  The length of the Cstr is greater than the length of the SubString
  section.  In this case the String is expanded before the assignment.
  </li>
</ul>

   As a result of the limitations imposed by the SubString
   constructor, a SubString will always be initialized with a String
   and the start and length will always be within bounds of that
   String.

 */
String SubString::operator =( const char *Cstr )
{
  size_t charsToCopy;

  if (Cstr && (charsToCopy = ::strlen( Cstr)) > 0) {
    size_t start = value->start();

    // create a copy region that is the correct size
    createCopySpace( start, charsToCopy );

    size_t cnt = 0;
    for (size_t i = start; i < start + charsToCopy; i++, cnt++) {
      value->pStr()->write(i, Cstr[ cnt ] );
    }
  }

  //
  // C/C++ allows chained assignment, so one can do something like
  //
  //    String a = b(2,4) = "abcd";
  //
  // So the string that results from the substring assignment should
  // be returned.
  //
  String retStr( *(value->pStr()) );

  return retStr;
} // operator = (Cstr)



/**
  operator =: Assign a String to a SubString.  For example:

<pre>
     String a("the quick brown fox jumped");
     String b("dog");

     a(16, 3) = b;
</pre>

    This function does the same thing as the C-string version
    of operator=.  Please see the comment associated with this
    function.

 */
String SubString::operator =( String& str )
{
  size_t charsToCopy;

  if ((charsToCopy = str.strlen()) > 0) {
    size_t start = value->start();

    // create a copy region that is the correct size
    createCopySpace(start, charsToCopy );

    if (value->pStr()) {
      size_t i, cnt;
      for (i = start, cnt = 0; i < start + charsToCopy; i++,cnt++) {
	value->pStr()->write(i, str[cnt]);
      }
    }
  }

  if (value->pStr())
    return *(value->pStr());
  else {
    String empty;

    return empty;
  }
} // operator = (String)


/**
 * operator = (SubStr &)
 *

   For example:
<pre>
      a(0, 4) = b(4, 4);
</pre>

 */
String SubString::operator =(const SubString& str )
{
  // convert to a SubStr = String
  *this = str.string();
  return (*this).string();
} // operator=



/**
   Return the String section defined by the SubString as a 
   String result.

   A SubString object consists of a String and a range (e.g., start
   index and length) which defines a string section.  In a sense, a
   SubString object is intent, not action.  The string() function
   performs the action by sectioning the SubString into a String.

   No bounds check is made here (for example, to assure that 
   subStrEnd is less than sLen) because this check is made
   when the SubString is constructed.

 */
String SubString::string() const
{
  String retVal;

  if (value->pStr()) {
    size_t sLen = value->pStr()->strlen();
    size_t start = value->start();
    size_t subStrLen = value->subStrLen();
    size_t subStrEnd = start + subStrLen;

    for (size_t i = start; i < subStrEnd; i++) {
      retVal += (*(value->pStr()))[i];
    }

  }
  return retVal;
} // SubString::string


/**
   Compare the SubString to a C string (e.g., const char *).  The function
   returns:

   <ul>
   <li>
   result == 0 when SubString is equal to the CString
   </li>
   <li>
   result &gt; 0 when the SubString is greater than the CString
   </li>
   <li>
   result &lt; 0 when the Substring is less than the Cstring
   </li>
   </ul>

   Notes:

   value->pStr() returns a pointer to the String which was used
   to initialize the SubString.  This String pointer is stored
   as part of the shared data.  Since compareTo is called when
   a SubString operation is being applied, we know that pStr()
   will is not NULL.  However, it IS possible that the string
   is empty.  In this case the String "const char *" operator
   will return null.

   If the strings are not the same length, the compareTo function
   acts as if the shorter string is extended with null characters
   until both strings are the same length.

   The compareTo function does not have to check the
   range of the SubString, since it was checked by the SubString
   constructor.
 */
int SubString::compareTo( const char *CStr )
{
  int result;
  const char *pThisCStr = *(value->pStr());
  if (pThisCStr == 0 && CStr == 0) {
    result = 0;
  } else if (pThisCStr == 0 && CStr != 0) {
    result = -1;
  } else if (pThisCStr != 0 && CStr == 0) {
    result = 1;
  } else { // both pThisCStr and CStr are non-NULL
    size_t subStrLen = value->subStrLen();
    size_t start = value->start();
    result = strncmp((pThisCStr + start), CStr, subStrLen);
    if (result == 0) {
      size_t CStrLen = strlen( CStr );
      if (CStrLen > subStrLen) {
	result = -1; // ThisCStr < CStr
      }
    }
  }
  return result;
} // compareTo for const char *


/**
   Compare the SubString to a String.  The function
   returns:

   <ul>
   <li>
   result == 0 when SubString is equal to the CString
   </li>
   <li>
   result &gt; 0 when the SubString is greater than the CString
   </li>
   <li>
   result &lt; 0 when the Substring is less than the Cstring
   </li>
   </ul>
*/
int SubString::compareTo( String &s )
{
  const char *pCStr = s;
  return compareTo( pCStr );
} // compareTo for String



/**
  Compare two SubStrings.

  As noted above, we know that the String is initialized in each
  case, although the Strings might still be null.  We also
  know that the start and length are in range, since this is
  checked in the constructor.

 */
int SubString::compareTo( SubString &subStr)
{
  int result;
  const char *pThisCStr = *(value->pStr());
  const char *pRhsCStr = *(subStr.value->pStr());

  // The Strings in both SubStrings are NULL
  if (pThisCStr == 0 && pRhsCStr == 0) {
    result = 0;
  }
  else if (pThisCStr != 0 && pRhsCStr == 0) {
    result = 1;
  }
  else if (pThisCStr == 0 && pRhsCStr != 0) {
    result = -1;
  }
  else {
    size_t thisLen = value->subStrLen();
    size_t rhsLen = subStr.value->subStrLen();
    size_t compareLen = thisLen;
    if (compareLen > rhsLen) {
      compareLen = rhsLen;
    }
    size_t thisStart = value->start();
    size_t rhsStart = subStr.value->start();

    result = strncmp( pThisCStr+thisStart,
		      pRhsCStr+rhsStart,
		      compareLen );
    if (result == 0) {
      if (thisLen < rhsLen) {
	result = -1;
      }
      else if (thisLen > rhsLen) {
	result = 1;
      }
    }
  }
  return result;
} // compareTo for SubString


//
// Global Operators
//

bool operator ==(const char *Cstr, SubString s)
{
  return (s == Cstr);
} // global operator ==

 
bool operator !=(const char *Cstr, SubString s)
{
  return (s != Cstr);
} // global operator !=


bool operator <=(const char *Cstr, SubString s)
{
  return (s >= Cstr);
} // global operator <=


bool operator >=(const char *Cstr, SubString s)
{
  return (s <= Cstr);
} // global operator >=

bool operator <(const char *Cstr, SubString s)
{
  return (s > Cstr);
} // global operator <

bool operator >(const char *Cstr, SubString s)
{
  return (s < Cstr);
} // global operator >
