
#ifndef SUBSTRING_H
#define SUBSTRING_H

#include <stdexcept>

/** \file

   <b>Copyright and Use</b>

   You may use this source code without limitation and without
   fee as long as you include:

<pre>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2001, 2002, 2003
</pre>

   This software is provided "as is", without any warrenty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.

   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>

 */

#include "RCObject.h"
#include "RCPtr.h"

class String;

/**
   This class supports the creation and manipulation of SubStrings,
   which are sections of Strings.  SubStrings are usually created by
   the "()" operator of the String class.  As a result, this class
   and the String class are closely related.

   The String class "()" operator takes two numeric arguments, the
   starting index and the length (String objects follow C/C++ indexing
   and start at 0).

   Examples:
<pre>
      String a("abcdefgh");
      String b;
      String tmp;

      b = a(0,4);        -- assignment of SubString to String
      
      a(0, 4) = "0123";  -- assignment of CString to SubString

      a(0, 4) = a(4, 4); -- assignment of SubString to SubString

      a(0, 4) = b;       -- assignment of String to SubString
</pre>

   Although a SubString is usually created by the String "()"
   operator, it can also be created explicitly.  For example:

<pre>
      String s = "the lazy dog";
      SubString t(s, 9, 3);

      t = "golden";
      std::cout << (const char *)s << std::endl;
</pre>

   Note that a SubString is a reference to a String.  So the
   assignment to the SubString alters the String.  The example
   above will write "the lazy golden" to std::cout.

   The SubString class suports relational operators.  This is faster
   than converting the SubString to a String and performing the
   relational operation on String operands.

   The SubString class does not support an argumentless constructor.
   This assures that all SubString objects are properly initialized
   with a String and a range (start index and length).  The starting
   index and length are checked in the constructor to assure that they
   are within the bounds of the String argument.  If they are not,
   an <tt>std::out_of_range</tt> exception will be thrown.

   There is no type cast supported for SubString to <tt>const char *</tt>
   although it might be nice to be able to do the following

<pre>
      String a = "morgan dreams of being fat";
      const char *sleep = a(0, 13);  <i>// Not supported!</i>
</pre>

   Unfortunately, there is no practical way to implement this.  The
   right-hand-side operation <tt>a(0, 13)</tt> creates a compiler
   generated String temporary.  This temporary will be
   deallocated after the statement containing the assignment.
   If the memory for the temporary were referenced from the 
   <tt>const char *</tt>, the <tt>const char *</tt> pointer would
   point to deallocated memory.

   There are some corner cases that I could not figure out how to
   protect against.  For example, it is possible to create a SubString
   object that references a deallocated string.  This is shown in the
   function <tt>makeSubString</tt> below:

<pre>
      <i>
      //
      // <b>Don't do this!</b>
      // This function will result in a SubString that references
      // a String which has been deallocated.
      //
      </i>
      SubString makeSubString(const char *Cstr, size_t start, size_t len)
      {
        String s( Cstr );
        SubString t(s, start, len);
        return t;  <i>// The String "s" will be deallocated on exit</i>
      }
</pre>

   For this function to work, the SubString constructor would have to
   increment the reference count of the String, so that the String
   would not be deallocated on exit (when its reference count reaches
   zero).  This could be done if an actual String, rather than a
   pointer to a String were used in the SubString shared data.

   Implementing the SubString in this way break the functionality of
   the SubString object.  Consider the code below:

<pre>
       String a = "12345678";
       a(3,0) = "abcd";
</pre>   

   If the SubString shared data contained a String, rather than a
   pointer to a string, the constructor for the SubString would
   increment the reference count for <i>a</i> to 2.  This would work
   well for the <tt>makeSubString</tt> function, but would break the
   assignment above.  

   If the SubString shared data contained a String, the assignment to
   this String would result in a unique copy being be made, since the
   String is shared (e.g., reference count > 1).  The C-string "abcd"
   would be inserted into the String in the SubString shared data
   (which is now unique).  The result will be that "a" is unchanged
   and the String we have assigned into will be deallocated when the
   SubString destructor is executed.

   In contrast to the String class, only a few SubString functions are
   inlined.  Many SubString functions reference the String class.  The
   String class is defined only as an abstract class here.  This is
   required by the fact that the String class also includes
   SubString.h.

   \author Ian Kaplan www.bearcave.com 
*/
class SubString 
{
private:

  class SharedData : public RCObject
  {
  private:
    String *pStr_;
    size_t start_;
    size_t subStrLen_;

  private:
    void copy(const SharedData &rhs);
    
  public:
    SharedData();
    SharedData(const SharedData &rhs) { copy( rhs ); }
    ~SharedData() {}

    void pStr( String *strRef ) { pStr_ = strRef; }
    String *pStr() { return pStr_; }
    
    void start(int start) { start_ = static_cast<size_t>(start); }
    size_t start() { return start_; }

    void subStrLen(int val) { subStrLen_ = static_cast<size_t>(val); }
    size_t subStrLen() { return subStrLen_; }

    size_t getRefCnt() { return refCnt(); }
  }; // class sharedData

  RCPtr<SharedData> value;

  void rangeCheck(String &s, const size_t start, const size_t len)
    throw(std::out_of_range);
  int compareTo( const char *CStr );
  int compareTo( String &s );
  int compareTo( SubString &subStr );
  void createCopySpace(size_t start, size_t charsToCopy);
  String string() const;

public:

  SubString(String &s, const size_t start, const size_t len) 
          throw(std::out_of_range);

  operator String() const;

  size_t getRefCnt() { return value->getRefCnt(); }

  size_t length() { return value->subStrLen(); }

  String operator +(const char *Cstr);      // String = SubString + Cstr
  String operator +(const SubString& sub);  // String = SubString + SubString
  String operator +(const String &str);     // String = SubString + String

  String operator =(const char *CStr );    // lhs SubString = Cstr
  String operator =(String& str );         // lhs SubString = String
  String operator =(const SubString& sub ); // lhs SubString = SubString

  // Relational operators
  bool operator ==(const char *Cstr ) { return (compareTo( Cstr ) == 0); }
  bool operator ==(String &s ) { return (compareTo( s ) == 0); }
  bool operator ==(SubString subStr) { return (compareTo(subStr) == 0); }

  bool operator !=(const char *Cstr ) { return (compareTo( Cstr ) != 0); }
  bool operator !=(String &s )  { return (compareTo( s ) != 0); }
  bool operator !=(SubString subStr) { return (compareTo(subStr) != 0); }

  bool operator <=(const char *Cstr ) { return (compareTo( Cstr ) <= 0); }
  bool operator <=(String &s )  { return (compareTo( s ) <= 0); }
  bool operator <=(SubString subStr) { return (compareTo(subStr) <= 0); }

  bool operator >=(const char *Cstr ) { return (compareTo( Cstr ) >= 0); }
  bool operator >=(String &s )  { return (compareTo( s ) >= 0); }
  bool operator >=(SubString subStr) { return (compareTo(subStr) >= 0); }

  bool operator <(const char *Cstr )  { return (compareTo( Cstr ) < 0); }
  bool operator <(String &s )   { return (compareTo( s ) < 0); }
  bool operator <(SubString subStr) { return (compareTo(subStr) < 0); }

  bool operator >(const char *Cstr )  { return (compareTo( Cstr ) > 0); }
  bool operator >(String &s )   { return (compareTo( s ) > 0); }
  bool operator >(SubString subStr) { return (compareTo(subStr) > 0); }
}; // SubString

//
// Global Operators
//
bool operator ==(const char *Cstr, SubString s);
bool operator !=(const char *Cstr, SubString s);
bool operator <=(const char *Cstr, SubString s);
bool operator >=(const char *Cstr, SubString s);
bool operator <(const char *Cstr, SubString s);
bool operator >(const char *Cstr, SubString s);


/** ----------------- SharedData methods ----------------- */


inline
SubString::SharedData::SharedData()
{
  pStr_ = 0;
  start_ = 0;
  subStrLen_ = 0;
}

inline
void SubString::SharedData::copy(const SharedData &rhs)
{
  pStr_ = rhs.pStr_;
  start_ = rhs.start_;
  subStrLen_ = subStrLen_;
} // copy


#include "String.h"

#endif

